// orders.c: Unit order processing for Bapple2
#include <stdio.h>
#include <stdlib.h>
#include <allegro.h>
#include "bapple2.h"
#include "bap2dat.h"
#include "globals.h"
#include "msg.h"
#include "units.h"
#include "map.h"
#include "comm.h"

void head_towards(int x1, int y1, int xoff1, int yoff1, int x2, int y2, int xoff2, int yoff2, int *direction);
unit_t *create_unit(signed char type, int x, int y, int direction, unsigned char owner);
void destroy_unit(unit_t *u);
void free_path_nodes(unit_t *u);
void destroy_all_units(void);
void issue_order(unit_t *u, char neworders);

void process_orders(unit_t *u);
void order_goto(unit_t *u);
void order_goto_closest_mbase(unit_t *u);
void order_guard(unit_t *u);
void order_random(unit_t *u);
void order_seek_player(unit_t *u);
void order_wait(unit_t *u);


extern fixed distance(signed int x1, signed int y1, signed int x2, signed int y2);
extern int create_projectile(float speed, signed char type, signed char team, player_t *ptarget, unit_t *utarget, int x, int y, int xoff, int yoff, int direction);
extern int play_sound_vol(SAMPLE *spl, int x, int y, int vol, int pan, int freq, int loop);
extern int FreeTile(int x, int y);
extern void find_closest_free_tile(int *dx, int *dy);
extern struct NODE *FindPath(long sx,long sy,long dx,long dy);
extern void fix_coords(int *x, int *y, signed int *xoff, signed int *yoff);

// From msg.c:
extern void add_msg(char type, char *text, ...);

// From map.c:
extern void destroy_window(wnd_t *wnd);

// From comm.c:
extern void send_unit_state(unit_t *u);

// Returns true if a unit exists at the specified coordinates
int unit_exists_at(int x, int y)
{
	thing_t *p;

	if (!mpos(unit_at, y, x))
		return 0;
	for (p = mpos(unit_at, y, x); p; p = p->next)
		if (p->type == THING_UNIT)
			return 1;

	return 0;
}

// Adds a thing to a location in the thing map
void add_thing(char type, void *ptr, int x, int y)
{
	thing_t *p;

	p = (thing_t *) malloc(sizeof(thing_t));
	p->type = type;
	p->ptr = ptr;
	p->next = mpos(unit_at, y, x);
	mpos(unit_at, y, x) = p;
}

// Removes a thing from a location in the thing map
// Returns true if successful, false otherwise
int remove_thing(char type, void *ptr, int x, int y)
{
	thing_t *p, *tmp, *bp;

	if (!mpos(unit_at, y, x))
		return 0;

	if (mpos(unit_at, y, x)->ptr == ptr)
	{
		tmp = mpos(unit_at, y, x)->next;
		free(mpos(unit_at, y, x));
		mpos(unit_at, y, x) = tmp;
		return 1;
	}

	bp = mpos(unit_at, y, x);
	for (p = mpos(unit_at, y, x)->next; p; p = p->next)
	{
		if (p->ptr == ptr)
		{
			bp->next = p->next;
			free(p);
			return 1;
		}
		bp = p;
	}

	return 0;
}

void process_orders(unit_t *u)
{
	if (u->carrier == -1)	// Only process orders when put on the ground
	switch (u->orders)
	{
		case ORDER_GUARD:  		order_guard(u); break;
        case ORDER_RANDOM: 		order_random(u); break;
        case ORDER_SEEK_PLAYER:	order_seek_player(u); break;
		case ORDER_WAIT:  		order_wait(u); break;
		case ORDER_GOTO:  		order_goto(u); break;
		case ORDER_GOTO_CLOSEST_MBASE:	order_goto_closest_mbase(u); break;
	}
}

void order_goto_closest_mbase(unit_t *u)
{
	int lx, ly;
	struct MBASE_NODE *mn;
	struct MBASE_NODE *mnode;
	fixed dist, d;
	thing_t *tmpthing;

	// Find the closest surviving enemy base
	dist = itofix(10000);
	for (mn = mbase; mn != NULL; mn = mn->next)
		if ((mn->health > 0) && (mn->owner != u->owner))
		{
			d = distance(u->x, u->y, mn->x + 2, mn->y + 2);
			if (d < dist)
			{
				dist = d;
				mnode = mn;
			}
		}
	// No base was found
	if (dist == itofix(10000))
	{
		issue_order(u, ORDER_GUARD);
		return;
	}

	// Send the unit to attack the found enemy base
	desired_terrain = unit_terrain[u->type];
	lx = mnode->x + 2;
	ly = mnode->y + 2;
	// Find a position to take
	find_closest_free_tile(&lx, &ly);

	// Calculate a path to that position

	// First their tile must not be marked in unit_at though!
	tmpthing = mpos(unit_at, u->y, u->x);
	mpos(unit_at, u->y, u->x) = 0;
	// Calculate the path and store it with the unit
	u->path = (struct NODE *) FindPath((long)lx,(long)ly,(long)u->x,(long)u->y);
	u->dx = lx;
	u->dy = ly;
	// Now the unit tile can be remarked as being occupied
	mpos(unit_at, u->y, u->x) = tmpthing;
	if (u->path != NULL)
	{
		u->OPEN = OPEN;
		u->CLOSED = CLOSED;
		issue_order(u, ORDER_GOTO);
	}
	else
	{
		issue_order(u, ORDER_GUARD);
	}
	return;
}

void order_guard(unit_t *u)
{
	int lx, ly;
	signed char lxoff, lyoff;
	int i;
	int old_dir;
	struct MBASE_NODE *mn;
	struct MBASE_NODE *mnode;
	fixed dist, d;
	unit_t *p;
	BITMAP *ubmp;

	// If we have a target
	if (u->target)
	{
		// Is the enemy within range? (and not being carried)
		if ((abs(u->x - u->target->x) < 5) && (abs(u->y - u->target->y) < 5) && (u->target->carrier == -1))
		{
			old_dir = u->targetdir;

			// Turn towards the enemy
			if ((u->direction == u->targetdir))
				head_towards(u->x, u->y, u->xoff, u->yoff, u->target->x, u->target->y, u->target->xoff, u->target->yoff, &(u->targetdir));

			// If we're facing the enemy and haven't shot for a while, shoot at em
			if (old_dir == u->targetdir)
				while (ms_time - u->last_shot_time > u->shot_rate)
				{
					ubmp = (BITMAP *) datafile[UNIT1_PCX + u->type].dat;
					lxoff = fixtoi(fcos(itofix(u->direction)) * (ubmp->w - 4));
					lyoff = fixtoi(fsin(itofix(u->direction)) * (ubmp->h - 4));
					create_projectile(4.0, PROJECTILE_TYPE_GREEN, PROJECTILE_TEAM_UNIT, 0, 0, u->x, u->y, (u->xoff + (ubmp->w / 2) + lxoff), (u->yoff + (ubmp->h / 2) + lyoff), u->direction);
					play_sound_vol((SAMPLE *) datafile[SND_SHOOT2].dat, u->x, u->y, 128, 128, 1000 + (rand() % 100) - 50, FALSE);
					u->last_shot_time += u->shot_rate;
				}
			return;
		}
		else
		{
			u->target = NULL;	// Targets that go out of range should be lost
			return;
		}
	}
	else
	// Look for enemy units within range to aquire as targets
	for (p = units; p; p = p->next)
	{
		// Ignore myself, friends, and units that are being carried
		if ((p == u) || (p->owner == u->owner) || (p->carrier != -1))
			continue;
		// Is the enemy within range?
		if ((abs(u->x - p->x) < 5) && (abs(u->y - p->y) < 5))
		{
			// Aquire this enemy as a target
			u->target = p;
			u->last_shot_time = ms_time;
			return;
		}
	}

	// If we're not really guarding, don't proceed with this function
	if (u->orders != ORDER_GUARD)
		return;

	// Look for enemy bases within range if no targets were found
	for (mnode = mbase; mnode != NULL; mnode = mnode->next)
	{
		// Ignore dead bases and friendly ones
		if ((mnode->owner == u->owner) || (mnode->health <= 0))
			continue;

		if ((abs(u->x - (mnode->x + 2)) < 10) && (abs(u->y - (mnode->y + 2)) < 10))
		{
			old_dir = u->targetdir;

			if ((u->direction == u->targetdir))
				head_towards(u->x, u->y, u->xoff, u->yoff, mnode->x + 2, mnode->y + 2, 15, 15, &(u->targetdir));

			if ((old_dir == u->targetdir) && (rand() % 50 == 0))
			{
				ubmp = (BITMAP *) datafile[UNIT1_PCX + u->type].dat;
				lxoff = fixtoi(fcos(itofix(u->direction)) * (ubmp->w));
				lyoff = fixtoi(fsin(itofix(u->direction)) * (ubmp->h));
				create_projectile(4.0, PROJECTILE_TYPE_GREEN, PROJECTILE_TEAM_UNIT, 0, 0, u->x, u->y, (u->xoff + (ubmp->w / 2) + lxoff), (u->yoff + (ubmp->h / 2) + lyoff), u->direction);
				play_sound_vol((SAMPLE *) datafile[SND_SHOOT2].dat, u->x, u->y, 128, 128, 1000 + (rand() % 100) - 50, FALSE);
			}
			return;
		}
	}

	// Don't let people stuck on islands try to go to enemy bases
	// Instead, have em shoot at the player :)
//	if ((unit_terrain[u->type] == TERRAIN_LAND) && (u->owner != player.id))
	if ((u->owner != player.id) && (player.health > 0))
	{
		// Turn towards the player
	    if ((u->direction == u->targetdir) && (rand() % 5 == 0))
	        head_towards(u->x, u->y, u->xoff, u->yoff, player.x, player.y, player.xoff, player.yoff, &(u->targetdir));
		// Is the player in range?
		if ((abs(u->x - player.x) < 5) && (abs(u->y - player.y) < 5))
		// Fire at that player!
	    if (ms_time - u->last_shot_time > u->shot_rate / 2)
	    {
			u->last_shot_time = ms_time;
			ubmp = (BITMAP *) datafile[UNIT1_PCX + u->type].dat;
			lxoff = fixtoi(fcos(itofix(u->direction)) * (ubmp->w - 4));
			lyoff = fixtoi(fsin(itofix(u->direction)) * (ubmp->h - 4));
//			lxoff = fixtoi(fmul(fcos(itofix(direction)), itofix(16)));
//			lyoff = fixtoi(fmul(fsin(itofix(direction)), itofix(16)));
	
			if (rand() % 20 == 0)	// Rarely fire off missiles  (used to be % 5)
			{
				i = create_projectile(2.0, PROJECTILE_TYPE_MISSILE, PROJECTILE_TEAM_UNIT, 0, 0, u->x, u->y, (u->xoff + 5 + lxoff), (u->yoff + 2 + lyoff), u->direction);
				proj[i].ptarget = &player;
				play_sound_vol((SAMPLE *) datafile[SND_SHOOT3].dat, u->x, u->y, 128, 128, 1000 + (rand() % 100) - 50, FALSE);
			}
	        else	// Usually, shoot the green things
			{
				create_projectile(4.0, PROJECTILE_TYPE_GREEN, PROJECTILE_TEAM_UNIT, 0, 0, u->x, u->y, (u->xoff + (ubmp->w / 2) + lxoff), (u->yoff + (ubmp->h / 2) + lyoff), u->direction);
				play_sound_vol((SAMPLE *) datafile[SND_SHOOT2].dat, u->x, u->y, 128, 128, 1000 + (rand() % 100) - 50, FALSE);
			}
	    }
		return;
	}
/*
	// No land-based units are to attempt travel to enemy bases!!!
	if (unit_terrain[u->type] == TERRAIN_LAND)
		return;

	// Only have the units attacking bases in single-player mode for now
	if (comm_mode != COMM_NONE)
		return;

	// Stagger the starting times of attacking units
	if (rand() % 100 == 0)
		return;

	// Send the unit after the closest surviving enemy main base
	issue_order(u, ORDER_GOTO_CLOSEST_MBASE); */
}

void order_random(unit_t *u)
{
    if (rand() % 4 == 0)
        u->direction += (rand() % 32) - 16;
    if (u->direction > 255)
        u->direction -= 256;
    else if (u->direction < 0)
        u->direction += 256;
	u->targetdir = u->direction;
}

void order_seek_player(unit_t *u)
{
    if ((u->direction == u->targetdir) && (rand() % 5 == 0))
        head_towards(u->x, u->y, u->xoff, u->yoff, player.x, player.y, player.xoff, player.yoff, &(u->targetdir));
    u->xoff += fixtoi(fmul(fcos(itofix(u->direction)), u->speed));
    u->yoff += fixtoi(fmul(fsin(itofix(u->direction)), u->speed));
}

void order_wait(unit_t *u)
{
	int lx, ly;
	struct NODE *tmppath;
	thing_t *tmpthing;

/*					if (rand() % 4 == 0)
                        unit->direction += (rand() % 32) - 16;
                    if (unit->direction > 255)
                        unit->direction -= 256;
                    else if (unit->direction < 0)
                        unit->direction += 256; */
//                    issue_order(unit, ORDER_GOTO);
//                    if (mpos(unit_at, unit->y, unit->x) == i)
//						if (FreeTile(unit->path->x, unit->path->y))
                    if (ms_time - u->wait_start_time > 500) // .5 seconds
//                    	issue_order(unit, unit->old_orders);
                    {
                        // Otherwise they gotta recalculate their path...
                        // Figure out what their final destination was
						if (!(u->path))
						{
		                    issue_order(u, ORDER_GUARD);
							return;
						}
                        while (u->path->Parent != NULL)
                            u->path = u->path->Parent;
                        lx = u->path->x;
                        ly = u->path->y;
						free_path_nodes(u);

                        // Now calculate a new path for em
                        desired_terrain = unit_terrain[u->type];
                        find_closest_free_tile(&lx, &ly);
//                        mpos(map, y, x) = 26;
//                        if (FreeTile(lx, ly))
//                        {
							tmpthing = mpos(unit_at, u->y, u->x);
							mpos(unit_at, u->y, u->x) = 0;
		                    desired_terrain = unit_terrain[u->type];
                            u->path = (struct NODE *) FindPath((long)lx,(long)ly,(long)u->x,(long)u->y);
							if (u->path)
							{
								u->OPEN = OPEN;
								u->CLOSED = CLOSED;
								u->dx = lx;
								u->dy = ly;
							}
							else
								u->dx = u->dy = -1;
                            mpos(unit_at, u->y, u->x) = tmpthing;
//                        }
						if (u->path)
		                    issue_order(u, ORDER_GOTO);
					}
}

void order_goto(unit_t *u)
{
	int i, j, lx, ly, x, y, r;
	struct NODE *Node;
	thing_t *tmpthing;

	i = 0;

                    if (u->path != NULL)
                    {
						// Check to see if they're sharing a tile with another unit (bad)
/*						j = mpos(unit_at, unit->y, unit->x);
						if ((j) && (j != i))
						{
							unit->wait_start_time = ms_time;
							issue_order(unit, ORDER_WAIT);
							break;
						} */
                        // If they've reached the next step in the path...
                        if (((u->x == u->path->x) && (u->y == u->path->y))
                            && (u->xoff == 15) && (u->yoff == 15))
                        {
							// If it's a friendly unit, uncover some of the map
							if (u->owner == player.id)
							{
								r = unit_vis_range[u->type];
								for (y = u->y - r; y <= u->y + r; y++)
									for (x = u->x - r; x <= u->x + r; x++)
										if (BETWEEN(x, 0, MAP_WIDTH) && BETWEEN(y, 0, MAP_HEIGHT))
											mpos(vis, y, x) = VIS_LIGHT;
								redraw_terrain_bmp = 1;
							}
                            // If that was the last step...
                            if (u->path->Parent == NULL)
                            {
								free_path_nodes(u);
								u->dx = u->dy = -1;
                                issue_order(u, ORDER_GUARD);
								send_unit_state(u);
                            }
                            else    // They need to advance to the next step
                            {
                                // Get the next step in the path
                                u->path = u->path->Parent;
                                // Check to see if anyone is in the way...
//                                desired_terrain = uniterrain[unit->type];
//                                if (!FreeTile(unit->path->x, unit->path->y))
                                if ((unit_exists_at(u->path->x, u->path->y)) ||
									(mpos(terrain, u->path->y, u->path->x) != unit_terrain[u->type]))
                                {
//                                	mpos(map, unit->path->y, unit->path->x)++;
//                                    mpos(map, unit->path->y, unit->path->x) %= NUM_TILES;

                                    // If they haven't been waiting, make em wait
                                    if (u->old_orders != ORDER_WAIT)
                                    {
                                        issue_order(u, ORDER_WAIT);
                                        u->wait_start_time = ms_time;
										send_unit_state(u);
                                        return;
                                    }
                                    // Otherwise they gotta recalculate their path...
                                    // Figure out what their final destination was
//                                    while (u->path->Parent != NULL)
//                                        u->path = u->path->Parent;
//                                    lx = u->path->x;
//                                    ly = u->path->y;
									lx = u->dx;
									ly = u->dy;
									free_path_nodes(u);

                                    // Now calculate a new path for em
                                    find_closest_free_tile(&lx, &ly);
                                    if (FreeTile(lx, ly))
                                    {
                                        tmpthing = mpos(unit_at, u->y, u->x);
                                        mpos(unit_at, u->y, u->x) = 0;
			                            desired_terrain = unit_terrain[u->type];
                                        u->path = (struct NODE *) FindPath((long)lx,(long)ly,(long)u->x,(long)u->y);
										if (u->path)
										{
											u->OPEN = OPEN;
											u->CLOSED = CLOSED;
											u->dx = lx;
											u->dy = ly;
	                                        mpos(unit_at, u->y, u->x) = tmpthing;
	                                        issue_order(u, ORDER_WAIT);
	                                        u->wait_start_time = ms_time;
											send_unit_state(u);
	                                        return;
										}
										else
											issue_order(u, ORDER_RANDOM);
                                    }
                                    else	// This should never happen
                                      issue_order(u, ORDER_RANDOM); // so spaz if it does :)
                                }
/*                                else
                                {
                                    map[unit->path->y][unit->path->x]++;
                                    map[unit->path->y][unit->path->x] %= NUM_TILES;
                                } */
                                if (u->path == NULL)
                                    issue_order(u, ORDER_WAIT);
//                                  issue_order(unit, ORDER_RANDOM);
                                else
								{
                                    head_towards(u->x, u->y, u->xoff, u->yoff, u->path->x, u->path->y, 15, 15, &(u->targetdir));
//                                map[unit->path->y][unit->path->x] = 2;
								// Experimental: Mark where the unit's about to go as busy
								// Note- this was EXTREMELY successful!!!!
//									mpos(unit_at, u->path->y, u->path->x) = 1;
								}
                            }
                        }
						// If we still have a path, proceed on that path
                        if (u->path != NULL)
						{
// moo!									mpos(unit_at, u->path->y, u->path->x) = tmpthing;
//							i = direction;
//							head_towards(x, y, xoff, yoff, path->x, path->y, 16, 16, &i);
//							i = targetdir;
//							head_towards(x, y, xoff, yoff, path->x, path->y, 15, 15, &i);
							if ((!(u->target)) && (u->direction == u->targetdir))
							{
//								xoff += fixtoi(fmul(fcos(itofix(direction)), speed));
//								yoff += fixtoi(fmul(fsin(itofix(direction)), speed));
	                            lx = u->x * 32 + u->xoff;
								ly = u->y * 32 + u->yoff;
	                            i = u->path->x * 32 + 15;
								j = u->path->y * 32 + 15;
	
								if (i > lx)
	                            	(u->xoff)++;
	                            if (i < lx)
	                            	(u->xoff)--;
								if (j > ly)
	                            	(u->yoff)++;
	                            if (j < ly)
	                            	(u->yoff)--;
	
								remove_thing(THING_UNIT, u, u->x, u->y);
								fix_coords(&(u->x), &(u->y), &(u->xoff), &(u->yoff));
								add_thing(THING_UNIT, u, u->x, u->y);
							}
						}
                    }
                    else	// path == NULL, The unit no longer has their path pointer, guard
					{
						// If the unit has destination coordinates though..
						if ((u->dx != -1) && (u->dy != -1))
						{
							// Calculate a path for em
							lx = u->dx;
							ly = u->dy;
							free_path_nodes(u);

							find_closest_free_tile(&lx, &ly);
							if (FreeTile(lx, ly))
							{
								tmpthing = mpos(unit_at, u->y, u->x);
								mpos(unit_at, u->y, u->x) = 0;
	                            desired_terrain = unit_terrain[u->type];
								u->path = (struct NODE *) FindPath((long)lx,(long)ly,(long)u->x,(long)u->y);
								if (u->path)
								{
									u->OPEN = OPEN;
									u->CLOSED = CLOSED;
									u->dx = lx;
									u->dy = ly;
									mpos(unit_at, u->y, u->x) = tmpthing;
									issue_order(u, ORDER_WAIT);
									u->wait_start_time = ms_time;
									send_unit_state(u);
									return;
								}
								else
									issue_order(u, ORDER_RANDOM);
							}
							else	// This should never happen
								issue_order(u, ORDER_RANDOM); // so spaz if it does :)
						}
						else	// Otherwise they don't deserve to have ORDER_GOTO
						{
	                        issue_order(u, ORDER_GUARD);
							send_unit_state(u);
						}
					}
	// A neat little hack to get things to stop and shoot at enemies while en route
	order_guard(u);
}

/* head_towards(): --------------------------------------------------------
 *   Given two sets of coordinates and a pointer to an object's direction,
 *   this function will turn the given object (located at the 1st set of
 *   coordinates) towards the destination (2nd set) coordinates.
 */
void head_towards(int x1, int y1, int xoff1, int yoff1, int x2, int y2, int xoff2, int yoff2, int *direction)
{
    fixed dx, dy, slope;
    signed int dir;

    dx = itofix(((x2 << 5) + xoff2) - ((x1 << 5) + xoff1));
    dy = itofix(((y2 << 5) + yoff2) - ((y1 << 5) + yoff1));
    slope = fdiv(dy, dx);
    dir = fixtoi(fatan(slope));
    if ((x2 < x1)) // && (player.xoff < unit[i].xoff))
        dir += 128;
    if ((x2 == x1)) // && (player.xoff == unit[i].xoff))
    {
        if (y1 < y2)
           dir = 64;
        else
           dir = -64;
    }

// Get the angle between 0 and 255!!!  (stop the infinite circling, aiee!!)
    while (dir < 0)
       dir += 256;
    dir %= 256;

    *direction = dir;
}

/* create_unit(): ---------------------------------------------------------
 *   Creates a unit at a given point with random properties.
 */
unit_t *create_unit(signed char type, int x, int y, int direction, unsigned char owner)
{
    unit_t *u, *p;
	int i, lx, ly, r;

/*	switch (type)	// old C++ crap
	{
		case UNIT_QUAD: u = new quad; break;
		case UNIT_BOAT: u = new boat; break;
		case UNIT_TANK: u = new tank; break;
		case UNIT_GUNBOAT: u = new gunboat; break;
		default: add_msg(MSG_COMPUTER, "Error- tried to create unknown unit type %d", type); return 0;
	} */

	u = (unit_t *) malloc(sizeof(unit_t));

	// Add the unit to the linked list
	u->next = NULL;		// always default the next pointer to null
	i = 0;
	for (p = units; p; p = p->next)
		i++;
//	printf("==+==> Constructing unit #%d\n", i + 1);
	u->next = units;
	units = u;

	// Give the unit a fresh ID number for comm purposes
	u->id = owner * 1000 + fresh_id++;

	// Initialize things that are common for all units
	u->carrier = -1;
    u->last_shot_time = ms_time;
    u->OPEN = u->CLOSED = u->path = NULL;
    u->old_orders = ORDER_GUARD;
    u->orders = ORDER_GUARD;
	u->target = NULL;
	u->dx = u->dy = -1;

	// Initialize unit-specific parameters
	switch (type)
	{
		case UNIT_QUAD:
		{
		    if (rand() % 15)
		        u->health = 200;
		    else
		        u->health = 400;
		    if (rand() % 15)
		        if (rand() % 4)
		            u->speed = ftofix(3.0);
		        else
		            u->speed = ftofix(4.0);
		    else
		        u->speed = ftofix(6.0);
		    u->shot_rate = 450;
			break;
		}
		case UNIT_BOAT:
		{
		    if (rand() % 15)
		        u->health = 100;
		    else
		        u->health = 200;
		    if (rand() % 15)
		        if (rand() % 4)
		            u->speed = ftofix(2.0);
		        else
		            u->speed = ftofix(3.0);
		    else
		        u->speed = ftofix(5.0);
		    u->shot_rate = 550;
			break;
		}
		case UNIT_TANK:
		{
		    if (rand() % 15)
		        u->health = 400;
		    else
		        u->health = 800;
		    if (rand() % 15)
		        if (rand() % 4)
		            u->speed = ftofix(3.0);
		        else
		            u->speed = ftofix(4.0);
		    else
		        u->speed = ftofix(6.0);
		    u->shot_rate = 500;
			break;
		}
		case UNIT_GUNBOAT:
		{
		    if (rand() % 15)
		        u->health = 300;
		    else
		        u->health = 600;
		    if (rand() % 15)
		        if (rand() % 4)
		            u->speed = ftofix(3.0);
		        else
		            u->speed = ftofix(4.0);
		    else
		        u->speed = ftofix(5.0);
		    u->shot_rate = 550;
			break;
		}
	}

    u->x = x;
    u->y = y;

    u->xoff = 15; //rand() % 32;
    u->yoff = 15; //rand() % 32;
    u->type = type;
    u->direction = (direction + (direction % 32)) % 256;
    u->targetdir = u->direction;
	u->owner = owner;
	u->maxhealth = u->health;

	// Add this unit to the thing map
	add_thing(THING_UNIT, u, u->x, u->y);

	// Make the unit shed some light on the map, if it's friendly
	if (u->owner == player.id)
	{
		r = unit_vis_range[u->type];
		for (ly = u->y - r; ly <= u->y + r; ly++)
			for (lx = u->x - r; lx <= u->x + r; lx++)
				if (BETWEEN(lx, 0, MAP_WIDTH) && BETWEEN(ly, 0, MAP_HEIGHT))
					mpos(vis, ly, lx) = VIS_LIGHT;
		redraw_terrain_bmp = 1;
	}

    return u;
}

/* issue_order(): ---------------------------------------------------------
 *   This issues an order to a unit.  The function was created so that
 *   the old orders of a unit could easily be remembered (important).
 *   Perhaps it could also be useful for bookkeeping and/or debugging...
 */
void issue_order(unit_t *u, char neworders)
{
    u->old_orders = u->orders;
    u->orders = neworders;

	// Special initialization or clean-up when switching to certain types of orders
	switch (u->orders)
	{
		case ORDER_GUARD:
		{
			u->dx = u->dy = -1;
			free_path_nodes(u);
			break;
		}
	}

//	send_unit_state(u);
}

/* free_path_nodes(): -----------------------------------------------------
 *   Frees all nodes in a unit's path - must be done on unit destruction!
 */
void free_path_nodes(unit_t *u)
{
	struct NODE *Node;
	int j = 0;

	if (u->OPEN)
		if (u->OPEN->NextNode)
		{
			Node = u->OPEN->NextNode;
			while (Node != NULL)
			{
				j++;
				free(Node);
				Node = Node->NextNode;
			}
		}
	if (u->CLOSED)
		if (u->CLOSED->NextNode)
		{
			Node = u->CLOSED->NextNode;
			while (Node != NULL)
			{
				j++;
				free(Node);
				Node = Node->NextNode;
			}
		}
	if (debug)
		add_msg(MSG_COMPUTER, "Freed %d nodes for unit #%d of player #%d", j, u->id, u->owner);
	u->path = u->OPEN = u->CLOSED = NULL;

	heapused -= (j * sizeof(struct NODE));
}

void destroy_unit(unit_t *u)
{
	char found = 0;
	int i;
	unit_t *p, *bu;
	wnd_t *w;
	char buf[40];

//	printf("==-==> Destroying a unit\n");

	// Tell the other players to destroy this unit
	if (comm_mode != COMM_NONE)
	{
		buf[0] = PACKET_UKILL;
		*(long *)(buf + 1) = u->id;
		for (i = 0; i < ((comm_mode == COMM_SERVER) ? active_enemies : 1); i++)
			Send(&buf[0], 1 + sizeof(long), enemy[i].addr);
	}

	// Remove this unit from the thing map
	remove_thing(THING_UNIT, u, u->x, u->y);

	// Free the path nodes
	free_path_nodes(u);

	// Terminate any missiles that were targeting this unit
	for (i = 0; i < active_projectiles; i++)
		if (proj[i].type != PROJ_NONEXISTENT)
			if (proj[i].utarget == u)
			{
				destroy_projectile(&proj[i]);
				create_explosion(SMALL_EXPLOSION, proj[i].x, proj[i].y, proj[i].xoff, proj[i].yoff);
			}

	// Destroy any windows that may be open with info about this unit
	for (w = wnd_list; w; w = w->next)
		if (w->data == u)
			destroy_window(w);

	// Release this unit from being targeted by other units
	for (p = units; p; p = p->next)
		if (p->target == u)
			p->target = NULL;

	// Remove this unit from the global linked list of units
	if (!units->next)		// if the only element in the list is the top one..
	{
		if (units == u)
		{
			units = NULL;
			found = 1;
		}
	}
	else
	{
		if (units == u)	// if we're removing the first of more than one node...
		{
			units = units->next;	// just bump the top node pointer to its child
			found = 1;
		}
		else  // otherwise we're removing a node somewhere down the list
		{
			for (bu = p = units; p; p = p->next)
			{
				if (p == u)	// we've found the node to be removed
				{
					bu->next = u->next;
					found = 1;
					break;
				}
				bu = p;
			}
		}
	}
//	if (!found)
//		printf("Unit was not found in the master unit linked list!\n");
}

void destroy_all_units(void)
{
	int i = 0;

	while ((units) && (i++ < 2000))
		destroy_unit(units);

	if (i >= 2000)
		add_msg(MSG_COMPUTER, "ERROR: Infinite loop while trying to destroy all units!  File: %s  Line: %d", __FILE__, __LINE__);
}
